#!/usr/bin/env python

import numpy as np
import pylab as plt
import sys
import copy

def get_model_perp():
    kr = 1.0
    ktheta = 0.2
    delta = 0.1
    m1 = 2.5
    m2 = 1.0

    model = ph_model(dim_k = 1,
                     dim_d = 3,
                     lat = [[0.0, 0.0, 1.0]],
                     crt = [[0.0, delta*0.5, 0.0],
                            [0.0, 0.0      , 0.5]],
                     mas = [m1, m2])
    
    model.set_spring(kr, ktheta, 0, 1, [ 0])
    model.set_spring(kr, ktheta, 0, 1, [-1])

    model.description="Model from Fig. 8b"
    model.short_description="model_from_fig_8b"

    return model

def get_model_para():    
    krS = 1.5
    krL = 1.0
    kthetaS = 0.24
    kthetaL = 0.18
    delta = 0.1
    m1 = 2.5
    m2 = 1.0

    model = ph_model(dim_k = 1,
                     dim_d = 3,
                     lat = [[0.0, 0.0, 1.0]],
                     crt = [[0.0, 0.0, delta*0.5],
                            [0.0, 0.0,       0.5]],
                     mas = [m1, m2])

    model.set_spring(krS, kthetaS, 0, 1, [ 0])
    model.set_spring(krL, kthetaL, 0, 1, [-1])

    model.description="Model from Fig. 8c"
    model.short_description="model_from_fig_8c"
    
    return model
    

def get_model_2d():
    kr = 1.0
    ktheta = 0.2
    delta = 0.1
    m1 = 2.5
    m2 = 1.0

    model = ph_model(dim_k = 2,
                     dim_d = 3,
                     lat = [[1.0, 0.0, 0.0],
                            [0.0, 1.0, 0.0]],
                     crt = [[0.0, 0.0, delta*0.5],
                            [0.5, 0.0,       0.0],
                            [0.0, 0.5,       0.0]],
                     mas = [m1, m2, m2])    

    model.set_spring(kr, ktheta, 0, 1, [ 0, 0])
    model.set_spring(kr, ktheta, 0, 2, [ 0, 0])
    model.set_spring(kr, ktheta, 0, 1, [-1, 0])
    model.set_spring(kr, ktheta, 0, 2, [ 0,-1])

    model.description="Model from Fig. 9"
    model.short_description="model_from_fig_9"

    return model

def get_model_perp_more_neigh():
    kr=1.0
    ktheta=0.2
    delta=0.1
    m1=2.5
    m2=1.0

    model = ph_model(dim_k = 1,
                     dim_d = 3,
                     lat = [[0.0, 0.0, 1.0]],
                     crt = [[0.0, delta*0.5, 0.0],
                            [0.0, 0.0      , 0.5]],
                     mas = [m1, m2])
    
    model.set_spring(kr, ktheta, 0, 1, [ 0])
    model.set_spring(kr, ktheta, 0, 1, [-1])
    
    model.set_spring(kr*0.035, ktheta*0.028, 0, 1, [ 1])
    model.set_spring(kr*0.035, ktheta*0.028, 0, 1, [-2])
    model.set_spring(kr*0.037, ktheta*0.025, 0, 1, [ 2])
    model.set_spring(kr*0.037, ktheta*0.025, 0, 1, [-3])

    model.description="Fig. 8b + more neighbors"
    model.short_description="model_from_fig_8b_more_neighbors"

    return model

def get_model_para_more_neigh():
    krS=1.5
    krL=1.0
    kthetaS=0.24
    kthetaL=0.18
    delta=0.1
    m1=2.5
    m2=1.0

    model = ph_model(dim_k = 1,
                     dim_d = 3,
                     lat = [[0.0, 0.0, 1.0]],
                     crt = [[0.0, 0.0, delta*0.5],
                            [0.0, 0.0,       0.5]],
                     mas = [m1, m2])

    model.set_spring(krS, kthetaS, 0, 1, [ 0])
    model.set_spring(krL, kthetaL, 0, 1, [-1])
    
    model.set_spring(krS*0.035, kthetaS*0.028, 0, 1, [ 1])
    model.set_spring(krS*0.035, kthetaS*0.028, 0, 1, [-2])
    model.set_spring(krS*0.037, kthetaS*0.025, 0, 1, [ 2])
    model.set_spring(krS*0.037, kthetaS*0.025, 0, 1, [-3])

    model.description="Fig. 8c + more neighbors"
    model.short_description="model_from_fig_8c_more_neighbors"
    
    return model
    


class ph_model(object):
    def __init__(self,dim_k,dim_d,lat,crt,mas):
        """Unlike in pythtb package, atom positions here are in cartesian coordinates."""
        
        self._dim_k=dim_k
#        self._dim_r=dim_r
        self._dim_d=dim_d
        self._lat=np.array(lat,dtype=float)
        self._crt=np.array(crt,dtype=float)
        self._natom=self._crt.shape[0]
        self._mas=np.array(mas,dtype=float)
        self._springs=[]
        
    def set_spring(self,k_stretch,k_bending,ind_i,ind_j,ind_R):
        self._springs.append([k_stretch,k_bending,int(ind_i),int(ind_j),np.array(ind_R)])

    def get_dynamical(self,kvec):
        mat=np.zeros((self._natom*self._dim_d,self._natom*self._dim_d),dtype=complex)

        for s in self._springs:
            ui=s[2]
            uj=s[3]

            crti=np.copy(self._crt[ui])
            crtj=np.copy(self._crt[uj])+np.einsum("ij,i->j",self._lat,s[4])

            rij=crtj-crti
            rij=rij/np.sqrt(np.dot(rij,rij))

            use_vec=s[4]
            
            for l in range(self._dim_d):
                for m in range(self._dim_d):
                    mat[ui+l*self._natom,ui+m*self._natom]+=(s[0]-s[1])*rij[l]*rij[m]/np.sqrt(self._mas[ui]*self._mas[ui])
                    mat[uj+l*self._natom,uj+m*self._natom]+=(s[0]-s[1])*rij[l]*rij[m]/np.sqrt(self._mas[uj]*self._mas[uj])

                    mat[ui+l*self._natom,uj+m*self._natom]-=(s[0]-s[1])*rij[l]*rij[m]*np.exp(( 1.0j)*np.dot(kvec,use_vec))/np.sqrt(self._mas[ui]*self._mas[uj])
                    mat[uj+l*self._natom,ui+m*self._natom]-=(s[0]-s[1])*rij[l]*rij[m]*np.exp((-1.0j)*np.dot(kvec,use_vec))/np.sqrt(self._mas[ui]*self._mas[uj])

                    
            for l in range(self._dim_d):
                mat[ui+l*self._natom,ui+l*self._natom]+=s[1]/np.sqrt(self._mas[ui]*self._mas[ui])
                mat[uj+l*self._natom,uj+l*self._natom]+=s[1]/np.sqrt(self._mas[uj]*self._mas[uj])
                    
                mat[ui+l*self._natom,uj+l*self._natom]-=s[1]*np.exp(( 1.0j)*np.dot(kvec,use_vec))/np.sqrt(self._mas[ui]*self._mas[uj])
                mat[uj+l*self._natom,ui+l*self._natom]-=s[1]*np.exp((-1.0j)*np.dot(kvec,use_vec))/np.sqrt(self._mas[ui]*self._mas[uj])


        return mat
                    
            
    def get_angular(self,evector):
        if self._dim_d!=3:
            print("Wrong dimensionality!")
            exit()

        ll=np.zeros((self._natom,self._dim_d))
        for a in range(self._natom):
            ll[a,0]+=2.0*np.real(evector[a+1*self._natom])*np.imag(evector[a+2*self._natom])
            ll[a,0]-=2.0*np.real(evector[a+2*self._natom])*np.imag(evector[a+1*self._natom])
            ll[a,1]+=2.0*np.real(evector[a+2*self._natom])*np.imag(evector[a+0*self._natom])
            ll[a,1]-=2.0*np.real(evector[a+0*self._natom])*np.imag(evector[a+2*self._natom])
            ll[a,2]+=2.0*np.real(evector[a+0*self._natom])*np.imag(evector[a+1*self._natom])
            ll[a,2]-=2.0*np.real(evector[a+1*self._natom])*np.imag(evector[a+0*self._natom])
            
        return ll



def main():

    np.set_printoptions(3,suppress=True)
    
    qpts=np.linspace(-0.5,0.5,501,endpoint=True)
    qpts=qpts*(2.0*np.pi)

    for get_model in [get_model_perp,
                      get_model_para,
                      get_model_perp_more_neigh,
                      get_model_para_more_neigh]:

        model=get_model()
        
        all_evals=[]
        all_ll=[]
        
        fig,ax=plt.subplots()
        for j in range(qpts.shape[0]):            
            
            mat=model.get_dynamical([qpts[j]])
            evals,eigs=np.linalg.eigh(mat)
            eigs=eigs.T
            all_evals.append(evals)

            ll=[]
            for i in range(evals.shape[0]):
                ll.append(model.get_angular(eigs[i]))
            all_ll.append(ll)

        all_evals=np.array(all_evals)
        all_ll=np.array(all_ll)
        # [qpt, branch, atom, cartesian]

        plot_ang(ax,qpts,all_evals,all_ll,model.description,True,200.0)
        ax.set_xlabel(r"q ($2 \pi / a$)") 

        print_info(qpts,all_evals,all_ll,71,model)

        fig.tight_layout()
        fig.savefig(model.short_description+".pdf")
        
        
    model=get_model_2d()

    qq0=np.linspace(-0.31, 0.43,101,endpoint=True)
    qq1=np.linspace( 0.22, 0.71,101,endpoint=True)
    qpts=np.transpose([qq0,qq1])
    qpts=qpts*(2.0*np.pi)
    
    all_evals=[]
    all_ll=[]
    fig,ax=plt.subplots()
    for j in range(qpts.shape[0]):            
            
        mat=model.get_dynamical(qpts[j])
        evals,eigs=np.linalg.eigh(mat)
        eigs=eigs.T
        all_evals.append(evals)

        ll=[]
        for i in range(evals.shape[0]):
            ll.append(model.get_angular(eigs[i]))
        all_ll.append(ll)
                
    all_evals=np.array(all_evals)
    all_ll=np.array(all_ll)

    plot_ang(ax,qpts,all_evals,all_ll,model.description,False,0.1)
    ax.set_xlabel(r"Trajectory in q-space")

    print_info(qpts,all_evals,all_ll,39,model)

    fig.tight_layout()
    fig.savefig(model.short_description+".pdf")



    print("Computing 2d histogram...")
    model=get_model_2d()
    nn=50
    tmp=np.linspace(-0.5,0.5,nn,endpoint=False)
    tmp=tmp*(2.0*np.pi)
    all_ll=[]
    fig,ax=plt.subplots()
    for qi in range(nn):
        for qj in range(nn):
            qpt=[tmp[qi],tmp[qj]]
            
            mat=model.get_dynamical(qpt)
            evals,eigs=np.linalg.eigh(mat)
            eigs=eigs.T

            for i in range(evals.shape[0]):
                all_ll.append(np.sum(model.get_angular(eigs[i]),axis=0))
            
    all_ll=np.array(all_ll)
    
    ax.hist2d(all_ll[:,0],all_ll[:,1],weights=np.sqrt(all_ll[:,0]**2+all_ll[:,1]**2),bins=31)
    
    ax.set_aspect(1.0)

    ax.set_xlabel(r"$L_x$")
    ax.set_ylabel(r"$L_y$")
    ax.set_title(model.description+" (histogram)")

    fig.tight_layout()
    
    fig.savefig(model.short_description+"__histogram.pdf")
    print("done!")
    

def plot_ang(ax,qpts,all_evals,all_ll,desc,useq=False,factor=0.1):

    ax.set_xlabel("Phonon wavevector trajectory")
    ax.set_title(desc)
    
    nq=qpts.shape[0]
    
    for i in range(all_evals.shape[1]):
        if useq is False:
            xval=np.linspace(0.0,1.0,nq,endpoint=True)
        else:
            xval=qpts/(2.0*np.pi)
        yval=np.sign(all_evals[:,i])*np.sqrt(np.abs(all_evals[:,i]))

        ax.plot(xval,yval,"k-",zorder=-100,lw=0.7)
        
        # [qpt, branch, atom, cartesian]

        # |total L|^2
        size=np.abs(np.sum(all_ll[:,i,:,0],axis=1))**2
        plot_line_thick(ax,xval,yval,size,"r",-20,factor)
        
        # |L_a|^2 summed over atoms a
        size=np.sum(np.abs(all_ll[:,i,:,0])**2,axis=1)
        plot_line_thick(ax,xval,yval,size,"b",-30,0.1)

def plot_line_thick(ax,xval,yval,size,c,zorder,factor):
    if np.max(np.abs(size))<1.0E-5:
        return
    ax.fill_between(xval,yval-size*factor,yval+size*factor,color=c,zorder=zorder,alpha=0.5,edgecolor="none")
        
def print_info(qpts,all_evals,all_ll,ind,model):
    print("\n"*2)
    print("*"*80)
    print("* ",model.description)
    print("*"*80)
    print()
    
    print("Lattice:")
    for i in range(model._lat.shape[0]):
        print("   #",i,"      ", model._lat[i])
    print("Atom positions:")
    for i in range(model._crt.shape[0]):
        print("   #", i,"     ", model._crt[i])
    print("Atom mass:")
    for i in range(model._mas.shape[0]):
        print("   #", i,"     ", model._mas[i])
    print()
    print("Springs:")
    for i in range(len(model._springs)):
        print("   #", i,"     From atom: ", model._springs[i][2]," to atom: ",model._springs[i][3]," shifted by: ",model._springs[i][4],"  stretch,bending = ",model._springs[i][0],",",model._springs[i][1])

    print()
    print("q-point coordinate: ",qpts[ind])
    for i in range(all_evals[ind].shape[0]):
        print("   Mode #",i)
        print("      Frequency: ",nicefy(np.sign(all_evals[ind,i])*np.sqrt(np.abs(all_evals[ind,i]))))
        print("         Angular momentum, total")
        print("                 Lx,Ly,Lz equals: ",nicefy(np.sum(all_ll[ind,i,:,0])),",",nicefy(np.sum(all_ll[ind,i,:,1])),",",nicefy(np.sum(all_ll[ind,i,:,2])))        
        print("         Angular momentum, per atom")
        # [qpt, branch, atom, cartesian]
        for j in range(all_ll[ind,i].shape[0]):
            print("            on atom #",j,"   Lx,Ly,Lz equals: ",nicefy(all_ll[ind,i,j,0]),",",nicefy(all_ll[ind,i,j,1]),",",nicefy(all_ll[ind,i,j,2]))

    print("*"*80)
    print("*"*80)
    print("\n"*2)

def nicefy(x):
    return str(round(x,3)).rjust(8)
    
main()







